/*******************************************************************************
** ATI Mach64 integrated PLL driver
********************************************************************************
**
   Copyright (C) 1998, 	Isaac Connor,
			Mathew Wilkins 
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
** 
********************************************************************************
** 
** $Log: mach64.c,v $
** Revision 1.7  1998/11/23 21:55:27  iconnor
** Corrected, Added License info to all Mach64 files.
** Added Me to maintainers list
**
** Revision 1.6  1998/11/19 02:43:04  marcus
** License change
**
** Revision 1.5  1998/10/28 05:05:33  brian
** Did clock portion of cmd --> *cmd check_mode change, and some
** uint/sint namespace fixing (still much to do on that in clock/)
**
** Revision 1.4  1998/10/25 18:07:40  marcus
** More namespace cleanups
**
** Revision 1.3  1998/10/04 03:17:15  brian
** Added private data hook to KGI_COMMON -- clean and recompile kgicon!
**
** Revision 1.2  1998/08/25 05:13:06  iconnor
** Removed include ggi/maintainers.h ... from mach64 source files
**
** Revision 1.1  1998/08/25 04:56:33  iconnor
** Initial import of the new Mach64 code.  Now does detection, and mode setting.
** This driver is semi-almost working.  Very little works.
**
** Revision 1.2  1998/08/23 01:50:58  iconnor
** Minor mods no effective code changes
**
** Revision 1.1.1.1  1998/08/22 03:35:58  iconnor
** Yo
**
** Revision 1.1  1998/07/07 08:01:52  taylorj
** *** empty log message ***
**
** Revision 1.1  1998/01/08 07:50:04  ursetto
** Initial check-in.
**
** 
**
*/

#define DRIVER_NAME     "Mach64 CT/VT/GT integrated PLL clock driver"
#define DRIVER_REV      "$Revision: 1.7 $"


#include <kgi/module.h>
#include "ATI/mach64.h"
#include "ATI/mach64_io.h"

#define CLK0_PLL 0x03

#include "pll.h"

/* xf86 recommends N2 [i.e. post-VCO or p] > 0 for stable clock */

static const struct pll_limits dclk_pll =
{
	14318180,                       /* Fref                         */
	{ 1, 1 },                       /* Factor                       */
	/* changed this! OK so far... */
	{ 135000000 , 270000000 },       /* FVCO range                   */
	{ 1, 3 },                       /* Post-VCO divider range       */
	{ 3, 129 },                     /* Dividend range               */           
	{ 3, 33 }                       /* Divisor range                */
};

static struct kgi_clock clock =
{
    
	"ATI Technologies Inc.",              /* Manufacturer         */
	"Mach64 CT/VT/GT integrated clock",    /* Model                */
	GM_ALL,                         /* Supported modes      */
	0,                              /* Special capabilities */
	{ 0x7FFF, 0x7FFF },             /* Maximum resolution   */
	NULL,				/* Private data hook	*/
	CLK_TYPE_PROG,                  /* Clock type           */
	{{
		0, 135000000,    	/* Clock range          */
	}}
};

/* An SDAC command word is as follows:
 * **      [0:6]   s->mul - 2
 * **      [8:12]  s->div - 2
 * **      [13:14] s->p
 * */

#define INVALID_COMMAND ((uint) -1)

static kgi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->mul - 2 < 128);
	ASSERT(s->div - 2 < 32);
	ASSERT(s->p < 4);
    
	return  (s->mul - 2) | ((s->div - 2) << 8) | (s->p << 13);
}

static void pll_set(struct kgi_display *dpy, kgi_uint cmd)
{
	kgiu8 hold;
       	kgiu32 data;
	printk("pll_set M=%d N=%d P=%d\n", cmd&0x7F, 
				(cmd>>8)&0x1F, (cmd>>13)&0x3);
    
    /* Set DCLK PLL parameters. */
	
	/* select PLL_REF_DIV=M */
	data = (1 << 9) | ((cmd&0x7F) << 16) | (2 << 10); 
	mach64_out32(Clock_Cntl, data, dpy);

	/* select VCLK0_POST_DIV=P */
	switch ((cmd >> 13 )& 3)
	{
		case 1:
		hold = 1; break;
		case 2:
		hold = 2; break;
		case 3:
		hold = 3; break;
		case 4:
		hold = 4; break;
		default: hold = 1;
	}
			
	data = (1 << 9) | (hold << 16) | (6 << 10); 
	mach64_out32(Clock_Cntl, data, dpy);
	
	/* select VCLK0_FB_DIV=N */
	data = (1 << 9) | (((cmd >> 8)&0x1F) << 16) | (7 << 10); 
	mach64_out32(Clock_Cntl, data, dpy);
	
	/*select VCLK_SRC_SEL=3  */
        data = (1 << 9) | (3 << 16) | (5 << 10);	
	mach64_out32(Clock_Cntl, data, dpy);
	
}

#include "pll.inc"                         
